#include <stdio.h>
#include <string.h>
#include "../uip/uip.h"
#include "dhcpc.h"
#include "../uip/timer.h"
#include "../uip/pt.h"
#include "../uip/uip_arp.h"
#include "../driver/CP2200.h"

// The DHCP client stores it's status in the following structure.
struct dhcpc_state {
    struct pt pt;              // Protothread of the DHCP client
    u8_t state;                 // Status, see DHCPC_STATE_xx values
    struct timer timer;        // Timer for waiting for repsonse
    u8_t retries;               // Counter for retries

    // the result:
    u16_t serverid[2];          // received ip address of the DHCP server
    u16_t lease_time[2];        // received lease time in seconds
    u16_t ipaddr[2];            // received ip address
    u16_t netmask[2];           // received network mask
    u16_t default_router[2];    // received default router
    u16_t dnsaddr[2];           // received dns address
};
static struct dhcpc_state status;

struct dhcp_msg {
    u8_t op, htype, hlen, hops;
    u8_t xid[4];
    u16_t secs, flags;
    u8_t ciaddr[4];
    u8_t yiaddr[4];
    u8_t siaddr[4];
    u8_t giaddr[4];
    u8_t chaddr[16];
    u8_t sname[64];
    u8_t file[128];
    u8_t options[312];
};

#define RETRY_TIMEOUT 2  // configurable in seconds

#define BOOTP_BROADCAST 0x8000

#define DHCP_REQUEST        1
#define DHCP_REPLY          2
#define DHCP_HTYPE_ETHERNET 1
#define DHCP_HLEN_ETHERNET  6
#define DHCP_MSG_LEN      236

#define DHCPDISCOVER  1
#define DHCPOFFER     2
#define DHCPREQUEST   3
#define DHCPDECLINE   4
#define DHCPACK       5
#define DHCPNAK       6
#define DHCPRELEASE   7

#define DHCP_OPTION_SUBNET_MASK   1
#define DHCP_OPTION_ROUTER        3
#define DHCP_OPTION_DNS_SERVER    6
#define DHCP_OPTION_REQ_IPADDR   50
#define DHCP_OPTION_LEASE_TIME   51
#define DHCP_OPTION_MSG_TYPE     53
#define DHCP_OPTION_SERVER_ID    54
#define DHCP_OPTION_REQ_LIST     55
#define DHCP_OPTION_END         255

#define DHCPC_STATE_INITIAL           0
#define DHCPC_STATE_SENT_DISCOVER     1
#define DHCPC_STATE_OFFER_RECEIVED    2
#define DHCPC_STATE_SENT_REQUEST      3
#define DHCPC_STATE_CONFIG_RECEIVED   4
#define DHCPC_STATE_NAK_RECEIVED      5
#define DHCPC_STATE_TIMEOUT           6
#define DHCPC_STATE_SENT_REFRESH      7
#define DHCPC_STATE_REFRESH_SUCCEEDED 8

static const u8_t xid[4] = {0xad, 0xde, 0x12, 0x23};
static const u8_t magic_cookie[4] = {99, 130, 83, 99};

/*---------------------------------------------------------------------------*/
static u8_t * add_msg_type(u8_t *optptr, u8_t type) {
    *optptr++ = DHCP_OPTION_MSG_TYPE;
    *optptr++ = 1;
    *optptr++ = type;
    return optptr;
}

/*---------------------------------------------------------------------------*/
static u8_t * add_server_id(u8_t *optptr) {
    *optptr++ = DHCP_OPTION_SERVER_ID;
    *optptr++ = 4;
    memcpy(optptr, status.serverid, 4);
    return optptr + 4;
}


/*---------------------------------------------------------------------------*/
static u8_t * add_req_ipaddr(u8_t *optptr) {
    *optptr++ = DHCP_OPTION_REQ_IPADDR;
    *optptr++ = 4;
    memcpy(optptr, status.ipaddr, 4);
    return optptr + 4;
}

/*---------------------------------------------------------------------------*/
static u8_t * add_req_options(u8_t *optptr) {
    *optptr++ = DHCP_OPTION_REQ_LIST;
    *optptr++ = 3;
    *optptr++ = DHCP_OPTION_SUBNET_MASK;
    *optptr++ = DHCP_OPTION_ROUTER;
    *optptr++ = DHCP_OPTION_DNS_SERVER;
    return optptr;
}

/*---------------------------------------------------------------------------*/
static u8_t * add_end(u8_t *optptr) {
    *optptr++ = DHCP_OPTION_END;
    return optptr;
}

/*---------------------------------------------------------------------------*/
static void create_msg(register struct dhcp_msg *m) {
    m->op = DHCP_REQUEST;
    m->htype = DHCP_HTYPE_ETHERNET;
    m->hlen = sizeof (uip_ethaddr.addr);
    m->hops = 0;
    memcpy(m->xid, xid, sizeof (m->xid));
    m->secs = status.retries*RETRY_TIMEOUT;
    m->flags = HTONS(BOOTP_BROADCAST);
    memcpy(m->ciaddr, uip_hostaddr, sizeof (m->ciaddr));
    memset(m->yiaddr, 0, sizeof (m->yiaddr));
    memset(m->siaddr, 0, sizeof (m->siaddr));
    memset(m->giaddr, 0, sizeof (m->giaddr));
    memcpy(m->chaddr, uip_ethaddr.addr, sizeof (uip_ethaddr.addr));
    memset(&m->chaddr[sizeof (uip_ethaddr.addr)], 0, sizeof (m->chaddr) - sizeof (uip_ethaddr.addr));
    memset(m->sname, 0, sizeof (m->sname));
    memset(m->file, 0, sizeof (m->file));
    memcpy(m->options, magic_cookie, sizeof (magic_cookie));
}

/*---------------------------------------------------------------------------*/
static void send_discover(void) {
    u8_t *end;
    struct dhcp_msg *m = (struct dhcp_msg *) uip_appdata;
    create_msg(m);
    end = add_msg_type(&m->options[4], DHCPDISCOVER);
    end = add_req_options(end);
    end = add_end(end);
    uip_send(uip_appdata, end - (u8_t *) uip_appdata);
}

/*---------------------------------------------------------------------------*/
static void send_request(void) {
    u8_t *end;
    struct dhcp_msg *m = (struct dhcp_msg *) uip_appdata;
    create_msg(m);
    end = add_msg_type(&m->options[4], DHCPREQUEST);
    end = add_server_id(end);
    end = add_req_ipaddr(end);
    end = add_end(end);
    uip_send(uip_appdata, end - (u8_t *) uip_appdata);
}

/*---------------------------------------------------------------------------*/
static u8_t
parse_options(u8_t *optptr, int len) {
    u8_t *end = optptr + len;
    u8_t type = 0;
    while (optptr < end) {
        switch (*optptr) {
            case DHCP_OPTION_SUBNET_MASK:
                memcpy(status.netmask, optptr + 2, 4);
                break;
            case DHCP_OPTION_ROUTER:
                memcpy(status.default_router, optptr + 2, 4);
                break;
            case DHCP_OPTION_DNS_SERVER:
                memcpy(status.dnsaddr, optptr + 2, 4);
                break;
            case DHCP_OPTION_MSG_TYPE:
                type = *(optptr + 2);
                break;
            case DHCP_OPTION_SERVER_ID:
                memcpy(status.serverid, optptr + 2, 4);
                break;
            case DHCP_OPTION_LEASE_TIME:
                memcpy(status.lease_time, optptr + 2, 4);
                break;
            case DHCP_OPTION_END:
                return type;
        }
        optptr += optptr[1] + 2;
    }
    return type;
}

/*---------------------------------------------------------------------------*/
static u8_t
parse_msg(void) {
    struct dhcp_msg *m = (struct dhcp_msg *) uip_appdata;
    if (m->op == DHCP_REPLY &&
            memcmp(m->xid, xid, sizeof (xid)) == 0 &&
            memcmp(m->chaddr, uip_ethaddr.addr, sizeof (uip_ethaddr.addr)) == 0) {
        memcpy(status.ipaddr, m->yiaddr, 4);
        return parse_options(&m->options[4], uip_datalen());
    }
    return 0;
}

/*---------------------------------------------------------------------------*/
static
PT_THREAD(handle_dhcp(void)) {
    PT_BEGIN(&status.pt);

    status.retries = 0;
    do {
        // send DHCP discover
        send_discover();
        status.state = DHCPC_STATE_SENT_DISCOVER;
        // wait 2 seconds for response
        timer_set(&status.timer, CLOCK_SECOND * RETRY_TIMEOUT); 
        PT_YIELD_UNTIL(&status.pt, uip_newdata() || timer_expired(&status.timer));
        if (uip_newdata()) {
            // if we received an answer, check if it was an OFFER
            if (parse_msg() == DHCPOFFER) {
                // received OFFER, now we need to send the REQUEST
                uip_flags &= ~UIP_NEWDATA;
                status.state = DHCPC_STATE_OFFER_RECEIVED;
                // send DHCP request
                send_request();
                status.state = DHCPC_STATE_SENT_REQUEST;
                // wait 4 seconds for response
                timer_set(&status.timer, CLOCK_SECOND * RETRY_TIMEOUT * 2);
                PT_YIELD_UNTIL(&status.pt, uip_newdata() || timer_expired(&status.timer));
                if (uip_newdata()) {
                    // if we received an answer, check if it was an ACK or NAK
                    if (parse_msg() == DHCPACK) {
                        uip_flags &= ~UIP_NEWDATA;
                        status.state = DHCPC_STATE_CONFIG_RECEIVED;
                        
                        // apply the received settings
                        uip_setnetmask(status.netmask);
                        uip_sethostaddr(status.ipaddr);
                        uip_setdraddr(status.default_router);
                        
                        // now we need to refresh the request repeatedly
                        // whenever half of the lease time has elapsed
                        while (1) {
                            timer_set(&status.timer, CLOCK_SECOND / 2); // 500ms
                            do {
                                PT_YIELD_UNTIL(&status.pt, timer_expired(&status.timer));
                                timer_reset(&status.timer);
                                status.lease_time[0]--;
                                if (status.lease_time[0] == 0) {
                                    status.lease_time[1]--;
                                }
                            }                               
                            while (status.lease_time[1] > 0 || status.lease_time[0] > 0);
                            
                            // send DHCP request
                            send_request();
                            status.state = DHCPC_STATE_SENT_REFRESH;
                            // wait 4 seconds for response
                            timer_set(&status.timer, CLOCK_SECOND * RETRY_TIMEOUT * 2); 
                            PT_YIELD_UNTIL(&status.pt, uip_newdata() || timer_expired(&status.timer));
                            if (uip_newdata()) {
                                // if we received an answer, check if it was an ACK or NAK
                                if (parse_msg() == DHCPACK) {
                                    uip_flags &= ~UIP_NEWDATA;
                                    status.state = DHCPC_STATE_REFRESH_SUCCEEDED;
                                }
                            } else {
                                status.state = DHCPC_STATE_TIMEOUT;
                            }
                        }
                        
                    } else if (parse_msg() == DHCPNAK) {
                        status.state = DHCPC_STATE_NAK_RECEIVED;
                    }
                } else {
                    status.state = DHCPC_STATE_TIMEOUT;
                }
            }
        } else {
            status.state = DHCPC_STATE_TIMEOUT;
        }
    }
    while (++status.retries < 30);
    
    // DHCP client failed, apply static settings
    uip_sethostaddr(config.ipaddr);
    uip_setnetmask(config.netmask);
    uip_setdraddr(config.gateway);

    // DHCP client ends here.
    // Enter an endless loop to prevent that the client restarts.
    while (1) {
	      PT_YIELD(&status.pt);
    }
    PT_END(&status.pt);
}

/*---------------------------------------------------------------------------*/
void
dhcpc_init(void) {
    uip_ipaddr_t addr;
    status.state = DHCPC_STATE_INITIAL;
    PT_INIT(&status.pt);

    if (config.dhcpenable) {
        // initially, the device IP-Address must be 0.0.0.0
        // otherwise the DHCP server would NAK the request
        uip_ipaddr(addr, 0, 0, 0, 0);
        uip_sethostaddr(addr);    

        // Destination is the broadcast address
        uip_ipaddr(addr, 255, 255, 255, 255);

        // Create the UDP connection
        struct uip_udp_conn *conn = uip_udp_new(&addr, HTONS(67));
        if (conn != NULL) {
            uip_udp_bind(conn, HTONS(68));
        } 
    }	
}

/*---------------------------------------------------------------------------*/
// this wrapper hides the protothread stuff of handle_dhcp()
void dhcpc_appcall(struct udp_application_state *app_state) {
    // Instead of storing the applications state in app_state, I decided to
    // use the static variable "state". This simplifies access for httpd and 
    // seriald (to display the current status), and it's Ok because dhcpc
    // support only one single connection any way.
    handle_dhcp();
}


/*----------------------------------------------------------------------------*/
// Output status as string

const char dhcpc_initial[]           PROGMEM="not started";
const char dhcpc_sent_discover[]     PROGMEM="sent discover";
const char dhcpc_offer[]             PROGMEM="offer received";
const char dhcpc_sent_request[]      PROGMEM="sent request";
const char dhcpc_config[]            PROGMEM="config received";
const char dhcpc_nak[]               PROGMEM="nak received";
const char dhcpc_timeout[]           PROGMEM="timeout";
const char dhcpc_sent_refresh[]      PROGMEM="sent refresh";
const char dhcpc_refresh_succeeded[] PROGMEM="refresh succeeded";

unsigned short get_dhcp_status(char* buffer) {
    int len=0;     
    switch (status.state) {
        case DHCPC_STATE_INITIAL:
            strcpy_P(buffer,dhcpc_initial);
            len=sizeof(dhcpc_initial)-1;
            break;
        case DHCPC_STATE_SENT_DISCOVER:
            strcpy_P(buffer,dhcpc_sent_discover);
            len=sizeof(dhcpc_sent_discover)-1;
            break;
        case DHCPC_STATE_OFFER_RECEIVED:
            strcpy_P(buffer,dhcpc_offer);
            len=sizeof(dhcpc_offer)-1;
            break;
        case DHCPC_STATE_SENT_REQUEST:
            strcpy_P(buffer,dhcpc_sent_request);
            len=sizeof(dhcpc_sent_request)-1;
            break;
        case DHCPC_STATE_CONFIG_RECEIVED:
            strcpy_P(buffer,dhcpc_config);
            len=sizeof(dhcpc_config)-1;
            break;
        case DHCPC_STATE_NAK_RECEIVED:
            strcpy_P(buffer,dhcpc_nak);
            len=sizeof(dhcpc_nak)-1;
            break;
        case DHCPC_STATE_TIMEOUT:
            strcpy_P(buffer,dhcpc_timeout);
            len=sizeof(dhcpc_timeout)-1;
            break;
        case DHCPC_STATE_SENT_REFRESH:
            strcpy_P(buffer,dhcpc_sent_refresh);
            len=sizeof(dhcpc_sent_refresh)-1;
            break;
        case DHCPC_STATE_REFRESH_SUCCEEDED:
            strcpy_P(buffer,dhcpc_refresh_succeeded);
            len=sizeof(dhcpc_refresh_succeeded)-1;
            break;
        default:
            // return a dummy because a null string is not allowed
            buffer[0]='?';
            len=1;
    }  
    return len;
}

