/*
    CP2201 driver.
    Major parts have been written by S.Perzborn.
    Modified by S.Frings.
 */

#include <inttypes.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "CP2200.h"
#include "../hw-layout.h"
#include "../uip/uip.h"
#include "../uip/uip_arp.h"

#define CPADDRH 0x21 // Current RX Packet Address High Byte page 65
#define CPADDRL 0x22 // Current RX Packet Address Low Byte page 65
#define CPINFOH 0x1D // Current RX Packet Information High Byte page 63
#define CPINFOL 0x1E // Current RX Packet Information Low Byte page 64
#define CPLENH 0x1F // Current RX Packet Length High Byte page 64
#define CPLENL 0x20 // Current RX Packet Length Low Byte page 64
#define CPTLB 0x1A // Current RX Packet TLB Number page 67
#define FLASHADDRH 0x69 // Flash Address Pointer High Byte page 76
#define FLASHADDRL 0x68 // Flash Address Pointer Low Byte page 76
#define FLASHAUTORD 0x05 // Flash AutoRead w/ increment page 77
#define FLASHDATA 0x06 // Flash Read/Write Data Register page 77
#define FLASHERASE 0x6A // Flash Erase page 77
#define FLASHKEY 0x67 // Flash Lock and Key page 76
#define FLASHSTA 0x7B // Flash Status page 75
#define CP2200_INT0 0x63 // Interrupt Status Register 0 (Self-Clearing) page 31
#define INT0EN 0x64 // Interrupt Enable Register 0 page 33
#define INT0RD 0x76 // Interrupt Status Register 0 (Read-Only) page 32
#define CP2200_INT1 0x7F // Interrupt Status Register 1 (Self-Clearing) page 34
#define INT1EN 0x7D // Interrupt Enable Register 1 page 36
#define INT1RD 0x7E // Interrupt Status Register 1 (Read-Only) page 35
#define IOPWR 0x70 // Port Input/Output Power page 45
#define MACADDR 0x0A // MAC Address Pointer page 79
#define MACDATAH 0x0B // MAC Data Register High Byte page 79
#define MACDATAL 0x0C // MAC Data Register Low Byte page 79
#define MACRW 0x0D // MAC Read/Write Initiate page 79
#define OSCPWR 0x7C // Oscillator Power page 46
#define PHYCF 0x79 // Physical Layer Configuration page 92
#define PHYCN 0x78 // Physical Layer Control page 91
#define PHYSTA 0x80 // Physical Layer Status page 93
#define RAMADDRH 0x08 // RAM Address Pointer High Byte page 24
#define RAMADDRL 0x09 // RAM Address Pointer Low Byte page 24
#define RAMRXDATA 0x02 // RXFIFO RAM Data Register page 24
#define RAMTXDATA 0x04 // TXBUFF RAM Data Register page 24
#define RSTEN 0x72 // Reset Enable Register page 42
#define RSTSTA 0x73 // Reset Source Status Register page 41
#define RXAUTORD 0x01 // RXFIFO AutoRead w/ increment page 62
#define RXCN 0x11 // Receive Control page 61
#define RXFIFOHEADH 0x17 // Receive Buffer Head Pointer High Byte page 71
#define RXFIFOHEADL 0x18 // Receive Buffer Head Pointer Low Byte page 71
#define RXFIFOSTA  0x5B // Receive Buffer Status page 72
#define RXFIFOTAILH 0x15 // Receive Buffer Tail Pointer High Byte page 71
#define RXFIFOTAILL 0x16 // Receive Buffer Tail Pointer Low Byte page 71
#define RXFILT 0x10 // Receive Filter Configuration
#define RXHASHH 0x0E // Receive Hash Table High Byte page 62
#define RXHASHL 0x0F // Receive Hash Table Low Byte page 63
#define RXSTA 0x12 // Receive Status page 61
#define SWRST 0x75 // Software Reset Register page 40
#define TLB0ADDRH 0x27 // TLB0 Address High Byte page 70
#define TLB0ADDRL 0x28 // TLB0 Address Low Byte page 70
#define TLB0INFOH 0x23 // TLB0 Information High Byte page 68
#define TLB0INFOL 0x24 // TLB0 Information Low Byte page 69
#define TLB0LENH 0x25 // TLB0 Length High Byte page 69
#define TLB0LENL 0x26 // TLB0 Length Low Byte page 70
#define TLB1ADDRH 0x2D // TLB1 Address High Byte page 70
#define TLB1ADDRL 0x2E // TLB1 Address Low Byte page 70
#define TLB1INFOH 0x29 // TLB1 Information High Byte page 68
#define TLB1INFOL 0x2A // TLB1 Information Low Byte page 69
#define TLB1LENH 0x2b // TLB1 Length High Byte page 69
#define TLB1LENL 0x2C // TLB1 Length Low Byte page 70
#define TLB2ADDRH 0x33 // TLB2 Address High Byte page 70
#define TLB2ADDRL 0x34 // TLB2 Address Low Byte page 70
#define TLB2INFOH 0x2F // TLB2 Information High Byte page 68
#define TLB2INFOL 0x30 // TLB2 Information Low Byte page 69
#define TLB2LENH 0x31 // TLB2 Length High Byte page 69
#define TLB2LENL 0x32 // TLB2 Length Low Byte page 70
#define TLB3ADDRH 0x39 // TLB3 Address High Byte page 70
#define TLB3ADDRL 0x3A // TLB3 Address Low Byte page 70
#define TLB3INFOH 0x35 // TLB3 Information High Byte page 68
#define TLB3INFOL 0x36 // TLB3 Information Low Byte page 69
#define TLB3LENH 0x37 // TLB3 Length High Byte page 69
#define TLB3LENL 0x38 // TLB3 Length Low Byte page 70
#define TLB4ADDRH 0x3F // TLB4 Address High Byte
#define TLB4ADDRL 0x40 // TLB4 Address Low Byte page 70
#define TLB4INFOH 0x3B // TLB4 Information High Byte page 68
#define TLB4INFOL 0x3C // TLB4 Information Low Byte page 69
#define TLB4LENH 0x3D // TLB4 Length High Byte page 69
#define TLB4LENL 0x3E // TLB4 Length Low Byte page 70
#define TLB5ADDRH 0x45 // TLB5 Address High Byte page 70
#define TLB5ADDRL 0x46 // TLB5 Address Low Byte page 70
#define TLB5INFOH 0x41 // TLB5 Information High Byte page 68
#define TLB5INFOL 0x42 // TLB5 Information Low Byte page 69
#define TLB5LENH 0x43 // TLB5 Length High Byte page 69
#define TLB5LENL 0x44 // TLB5 Length Low Byte page 70
#define TLB6ADDRH 0x4B // TLB6 Address High Byte page 70
#define TLB6ADDRL 0x4C // TLB6 Address Low Byte page 70
#define TLB6INFOH 0x47 // TLB6 Information High Byte page 68
#define TLB6INFOL 0x48 // TLB6 Information Low Byte page 69
#define TLB6LENH 0x49 // TLB6 Length High Byte page 69
#define TLB6LENL 0x4A // TLB6 Length Low Byte page 70
#define TLB7ADDRH 0x51 // TLB7 Address High Byte page 70
#define TLB7ADDRL 0x52 // TLB7 Address Low Byte page 70
#define TLB7INFOH 0x4D // TLB7 Information High Byte page 68
#define TLB7INFOL 0x4E // TLB7 Information Low Byte page 69
#define TLB7LENH 0x4F // TLB7 Length High Byte page 69
#define TLB7LENL 0x50 // TLB7 Length Low Byte page 70
#define TLBVALID 0x1C // TLB Valid Indicators page 68
#define TXAUTOWR 0x03 // Transmit Data AutoWrite page 53
#define TXBUSY 0x54 // Transmit Busy Indicator page 51
#define TXCN 0x53 // Transmit Control page 51
#define TXENDH 0x57 // Transmit Data Ending Address High Byte page 53
#define TXENDL 0x58 // Transmit Data Ending Address Low Byte
#define TXPAUSEH 0x55 // Transmit Pause High Byte page 52
#define TXPAUSEL 0x56 // Transmit Pause Low Byte page 52
#define TXPWR 0x7A // Transmitter Power page 46
#define TXSTA0 0x62 // Transmit Status Vector 0 page 57
#define TXSTA1 0x61 // Transmit Status Vector 1 page 56
#define TXSTA2 0x60 // Transmit Status Vector 2 page 56
#define TXSTA3 0x5F // Transmit Status Vector 3 page 55
#define TXSTA4 0x5E // Transmit Status Vector 4 page 55
#define TXSTA5 0x5D // Transmit Status Vector 5 page 54
#define TXSTA6 0x5C // Transmit Status Vector 6 page 54
#define TXSTARTH 0x59 // Transmit Data Starting Address High Byte page 52
#define TXSTARTL 0x5A // Transmit Data Starting Address Low Byte page 52
#define VDMCN 0x13 // VDD Monitor Control Register page 39

// Indirect MAC registers
#define MACCN  0x00 // MAC Control. Used to enable reception and other options.
#define MACCF 0x01 // MAC Configuration. Used to configure padding options and other settings.
#define IPGT 0x02 // Back-to-Back Interpacket Delay. Sets the Back-to-Back Interpacket Delay.
#define IPGR 0x03 // Non-Back-to-Back Interpacket Delay. Sets the Non-Back-to-Back Interpacket Delay.
#define CWMAXR 0x04 // Collision Window and Maximum Retransmit. Sets the collision window size and the maximum number of retransmits allowed.
#define MAXLEN 0x05 // Maximum Frame Length. Sets the maximum receive frame length.
#define MACAD0 0x10 // MAC Address 
#define MACAD1 0x11 // MAC Address 
#define MACAD2 0x12 // MAC Address 

// bits in PHYCF
#define SMSQ 0x80
#define LINKINTG 0x40
#define JABBER 0x20
#define AUTONEG 0x10
#define ADPAUSE 0x04
#define AUTOPOL 0x02
#define REVPOL 0x01

// bits in PHYCN
#define PHYEN 0x80
#define PHY_TXEN 0x40
#define PHY_RXEN 0x20
#define DPLXMD 0x10
#define LBMD 0x08
#define LPRFAULT 0x04
#define POLREV 0x02
#define LINKSTA 0x01

// bits in INT1
#define WAKEINT 0x20
#define LINKINT 0x10
#define JABINT 0x08
#define RFINT 0x02
#define ANCINT 0x01


// the configuration from flash memory
struct flashconfig config;

//------------------------------------------------
// Lookup table to swap bit order MSB<->LSB
// This table is in RAM for performance reason

#ifdef SWAP_ADBUS
static uint8_t swap[256] = {
   0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,
   0x10,0x90,0x50,0xd0,0x30,0xb0,0x70,0xf0,
   0x08,0x88,0x48,0xc8,0x28,0xa8,0x68,0xe8,
   0x18,0x98,0x58,0xd8,0x38,0xb8,0x78,0xf8,
   0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,
   0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4,
   0x0c,0x8c,0x4c,0xcc,0x2c,0xac,0x6c,0xec,
   0x1c,0x9c,0x5c,0xdc,0x3c,0xbc,0x7c,0xfc,
   0x02,0x82,0x42,0xc2,0x22,0xa2,0x62,0xe2,
   0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2,
   0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,
   0x1a,0x9a,0x5a,0xda,0x3a,0xba,0x7a,0xfa,
   0x06,0x86,0x46,0xc6,0x26,0xa6,0x66,0xe6,
   0x16,0x96,0x56,0xd6,0x36,0xb6,0x76,0xf6,
   0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,
   0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe,
   0x01,0x81,0x41,0xc1,0x21,0xa1,0x61,0xe1,
   0x11,0x91,0x51,0xd1,0x31,0xb1,0x71,0xf1,
   0x09,0x89,0x49,0xc9,0x29,0xa9,0x69,0xe9,
   0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9,
   0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,
   0x15,0x95,0x55,0xd5,0x35,0xb5,0x75,0xf5,
   0x0d,0x8d,0x4d,0xcd,0x2d,0xad,0x6d,0xed,
   0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,0x7d,0xfd,
   0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,
   0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3,
   0x0b,0x8b,0x4b,0xcb,0x2b,0xab,0x6b,0xeb,
   0x1b,0x9b,0x5b,0xdb,0x3b,0xbb,0x7b,0xfb,
   0x07,0x87,0x47,0xc7,0x27,0xa7,0x67,0xe7,
   0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7,
   0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,
   0x1f,0x9f,0x5f,0xdf,0x3f,0xbf,0x7f,0xff};
#endif
   
//-------------------------------------------------------------
// Wait-states for parallel interface to the ethernet controller

#if F_CPU <= 20000000
    // up to 20 Mhz, one clock cycle is at least 50ns 
    #define wait50ns_minus_1_clock()  { }
    #define wait150ns_plus_2_clock()  { asm volatile ("nop\n nop\n nop\n nop\n nop"); }
#elif F_CPU <= 40000000  
    // 21 Mhz - 40 Mhz, one clock cycle is at least 25ns
    #define wait50ns_minus_1_clock()  { asm volatile ("nop"); }
    #define wait150ns_plus_2_clock()  { asm volatile ("nop\n nop\n nop\n nop\n nop\n nop\n nop\n nop"); }  
#else
    #error The ethernet driver has been designed for max 40 Mhz.
#endif

//-------------------------------------------------------------

void write_CP2200(int adr, char value)
{
        nCSETH(0); // clear nCS
        ALE(1); // set ALE
        ADBUS_output(); // AD0..7 output
        #ifdef SWAP_ADBUS
            ADBUS_write(swap[adr]); // output address
        #else
            ADBUS_write(adr); // output address
        #endif
   	    wait50ns_minus_1_clock();
        ALE(0); // clear ALE
   	    wait50ns_minus_1_clock();
        nWR(0); // clear nWR
        #ifdef SWAP_ADBUS
            ADBUS_write(swap[(uint8_t)value]); // output address
        #else
           ADBUS_write(value); // output data
        #endif
        wait50ns_minus_1_clock();
        nWR(1); // set nWR
        wait50ns_minus_1_clock();
        nCSETH(1); // set nCS
        ADBUS_input(); // AD0..7 Tri-state
}


u8_t read_CP2200(int adr)
{
        char value;
        nCSETH(0); // clear nCS
        ALE(1); // set ALE
        ADBUS_output(); // AD0..7 output
        #ifdef SWAP_ADBUS
            ADBUS_write(swap[adr]); // output address
        #else
            ADBUS_write(adr); // output address
        #endif
        wait50ns_minus_1_clock();
        ALE(0); // clear ALE
        wait50ns_minus_1_clock();
        ADBUS_input(); // AD0..7 input
        nRD(0); // clear nRD
        wait150ns_plus_2_clock();
        #ifdef SWAP_ADBUS
            value = swap[ADBUS_read()]; // input D0..7
        #else
            value = ADBUS_read(); // input D0..7
        #endif        
        nRD(1); // set nRD
        nCSETH(1); // set nCS
        return(value);
}


// ---------------------------------------------
unsigned char CP2200_ReadTXBuffer(unsigned int addr)
{
        write_CP2200(RAMADDRL,addr & 0xFF);
        write_CP2200(RAMADDRH,addr >> 8);
        return (read_CP2200(RAMTXDATA));
}

// ---------------------------------------------
void CP2200_WriteTXBuffer(unsigned int addr, unsigned char value)
{
        write_CP2200(RAMADDRL,addr & 0xFF);
        write_CP2200(RAMADDRH,addr >> 8);
        write_CP2200(RAMTXDATA,value);
}

// ---------------------------------------------
unsigned char CP2200_ReadRXBuffer(unsigned int addr)
{
        write_CP2200(RAMADDRL, (char) (addr & 0xFF));
        write_CP2200(RAMADDRH,(char) (addr >> 8));
        return (read_CP2200(RAMRXDATA));
}

// ---------------------------------------------
void CP2200_WriteRXBuffer(unsigned int addr, unsigned char value)
{
        write_CP2200(RAMADDRL,addr & 0xFF);
        write_CP2200(RAMADDRH,addr >> 8);
        write_CP2200(RAMRXDATA,value);
}

// ---------------------------------------------
u16_t CP2200_ReadPacket(void)
{
        u16_t packetlen;
        u16_t i;
        packetlen = read_CP2200(CPLENH);
        packetlen = packetlen << 8;
        packetlen |= read_CP2200(CPLENL);
        if (packetlen>UIP_BUFSIZE) {
          packetlen=UIP_BUFSIZE;
        }
        for (i=0;i<packetlen;i++)
        {
                *(uip_buf+i) = read_CP2200(RXAUTORD);
        }
        return(packetlen);
}

// ---------------------------------------------
u16_t network_device_read(void)
{
        u16_t  packetlen=0;
        int cp_bufptr; // Zeiger auf Paket Puffer
        char status = read_CP2200(CP2200_INT0);     
        
        if ((status & 0x02) == 0x02)
        {
                write_CP2200(RXCN,0x01);
                write_CP2200(RXCN,0x00);
        }
                   
        // Paket im Empfangspuffer?
        if (((read_CP2200(CPINFOH) &0x80) != 0) && (read_CP2200(CPINFOL) &0x80) != 0)
        {
                // Zeiger auf Paket Puffer merken
                cp_bufptr = read_CP2200(RXFIFOHEADH);
                cp_bufptr = cp_bufptr << 8;
                cp_bufptr |= read_CP2200(RXFIFOHEADL);             
                packetlen = CP2200_ReadPacket();
                write_CP2200(RXCN,0x06); // discharge package
        }        
        return(packetlen);
}

// ---------------------------------------------
void CP2200_WritePacket(void)
{
        //unsigned int len;
        unsigned int addr;
        unsigned int t;
        // wait for previous packet complete
        while (read_CP2200(TXBUSY) != 0x00);
        // set transmit buffer pointer
        write_CP2200(TXSTARTH,0x00);
        write_CP2200(TXSTARTL,0x00);
        
        addr=0;
        for (t=0;t < UIP_LLH_LEN;t++)
        {
                CP2200_WriteTXBuffer(addr++, *(uip_buf + t));
        }
        
        if(uip_len <= UIP_LLH_LEN + UIP_TCPIP_HLEN)
        {
                for (t=UIP_LLH_LEN;t < uip_len;t++)
                {
                        CP2200_WriteTXBuffer(addr++, *(uip_buf + t));
                }
        }
        else
        {
                for (t=UIP_LLH_LEN;t < UIP_TCPIP_HLEN+UIP_LLH_LEN;t++)
                {
                        CP2200_WriteTXBuffer(addr++, *(uip_buf + t));
                }
                //hwsend(uip_appdata, uip_len - UIP_TCPIP_HLEN - UIP_LLH_LEN);
                for (t=0;t < uip_len - UIP_TCPIP_HLEN - UIP_LLH_LEN;t++)
                {
                        CP2200_WriteTXBuffer(addr++, *(((unsigned char *)uip_appdata) + t));
                }
        }

        // set packet end
        addr--;
        write_CP2200(TXENDH,(char) (addr >> 8));
        write_CP2200(TXENDL,(char) (addr & 0xFF));
        // set packet start
        write_CP2200(TXSTARTH,0x00);
        write_CP2200(TXSTARTL,0x00);
        // send packet
        write_CP2200(TXCN,0x01);
}


void network_device_send(void)
{
        CP2200_WritePacket();
}



void cpWriteMac(char adr,int value)
{
        write_CP2200(MACADDR,adr);
        write_CP2200(MACDATAH,value>>8);
        write_CP2200(MACDATAL,value);
        write_CP2200(MACRW,1);
}

// Initialize CP2200 so the link can come up
void init_PHY(void) {
    // Enable link LED and pull-ups
    write_CP2200(IOPWR,0x0C);    
    // init PHY in full duplex mode 
    write_CP2200(PHYCN,0x00);
    write_CP2200(PHYCN,DPLXMD);
    write_CP2200(PHYCF,SMSQ|LINKINTG|JABBER|AUTOPOL);
    write_CP2200(PHYCN,DPLXMD|PHYEN); 
    _delay_ms(2);
    // enable transmitter and receiver
    write_CP2200(PHYCN,DPLXMD|PHYEN|PHY_TXEN|PHY_RXEN);
    _delay_ms(100);
}

// Check if the link is up (link led is on)
uint8_t isLinkUp(void) {
    return (read_CP2200(PHYCN) & LINKSTA);
}

// Reset the ethernet controller
void eth_RESET(void) {
    #ifdef nRSTETH
        nRSTETH(0);
        _delay_ms(10);
        nRSTETH(1);
    #else
        write_CP2200(SWRST,4);
    #endif    
    _delay_ms(100);
}

// Initialize hte MAC address (after link comes up)
void init_MAC(void) {
        unsigned int tmp16;
               
        //MAC INIT
        cpWriteMac(MACCF,0x40B3); // full duplex
        cpWriteMac(IPGT,0x0015);  // full duplex
        cpWriteMac(MACCF,0x40B3);
        cpWriteMac(MAXLEN,0x05EE);
        write_CP2200(FLASHADDRL,0xFA);
        write_CP2200(FLASHADDRH,0x1F);
        tmp16 = 0x0000;
        tmp16 = read_CP2200(FLASHAUTORD);
        tmp16 |=read_CP2200(FLASHAUTORD)<<8;
        cpWriteMac(MACAD2,tmp16);
        tmp16 = 0x0000;
        tmp16 = read_CP2200(FLASHAUTORD);
        tmp16 |= read_CP2200(FLASHAUTORD)<<8;
        cpWriteMac(MACAD1,tmp16);
        tmp16 = 0x0000;
        tmp16 = read_CP2200(FLASHAUTORD);
        tmp16 |= read_CP2200(FLASHAUTORD)<<8;
        cpWriteMac(MACAD0,tmp16);
        cpWriteMac(MACCN,0x01);

        // Enable Packet Receive and Receive FIFO full interrupts
        write_CP2200(INT0EN,0x03);
        write_CP2200(INT1EN,0x00);
        
        // clear interrupt status
        read_CP2200(CP2200_INT0);
        read_CP2200(CP2200_INT1);
        struct uip_eth_addr eaddr;

        // configure uIP with own mac address (get from CP2200 flash memory)
        // added by Stefan Frings
        write_CP2200(FLASHADDRL,0xFA);
        write_CP2200(FLASHADDRH,0x1F);
        for (int i=0;i<6;i++)
        {
            eaddr.addr[i]=read_CP2200(FLASHAUTORD); // mac address from CP2200
        }
        // set mac address of this modules
        uip_setethaddr(eaddr);
}


// write a single byte into the flash.
// the flash must be erased before that.

void writeflash(u16_t adr, u8_t data) {
    // unlock the flash write protection
    cli();
    write_CP2200(FLASHKEY,0xA5);
    write_CP2200(FLASHKEY,0xF1);
    // select the write address
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    write_CP2200(FLASHDATA,data);
    sei();
    // wait until writing finishes
    u8_t flash_status;
    do {
        flash_status=read_CP2200(FLASHSTA);
    } while ((flash_status & 8)>0);
}



// Read a single character from the flash memory at the given adress.
// code 255 is converted to code 0.

u8_t readFlashChar(u16_t adr) {
    char c;
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    c=read_CP2200(FLASHAUTORD);
    if (c==255)
        c=0;
    return c;
}


// Read a String from the flash memory at the given adress.

void readFlashString(char* buffer, u16_t adr) {
    char c;
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    do {
        c=read_CP2200(FLASHAUTORD);
        // Replace ascii 255 by ascii 0.
        // Epty flash cells have that value
        if (c==255)
          c=0;
        buffer[0]=c;
        buffer++;
    }
    while (c!=0);
}

// Write IP configuration into flash.
void writeFlashConfig() {
    u16_t adr=FLASHCONFIG;
    // unlock the flash write protection
    cli();
    write_CP2200(FLASHKEY,0xA5);
    write_CP2200(FLASHKEY,0xF1);
    // erase the page with config settings
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    write_CP2200(FLASHERASE,0x01);
    sei();
    // wait until erasing finishes
    u8_t flash_status;
    do {
        flash_status=read_CP2200(FLASHSTA);
    } while ((flash_status & 8)>0);
    u8_t* ptr= (u8_t*) &config;
    for (int i=0; i<sizeof(config); i++) {
        writeflash(adr++,ptr[i]);
    }
}


// Write IP configuration from flash.
void readFlashConfig(void) {
    u16_t adr=FLASHCONFIG;
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    u8_t* ptr=(u8_t*)&config;
    for (int i=0; i<sizeof(config); i++) {
        ptr[i]=read_CP2200(FLASHAUTORD);
    }
}

// Write String to the flash memory at the given adress.
// The string must be 0-512 characters long and it must start at a 512-byte boundary.

void writeFlashString(char* string, u16_t adr) {
    u8_t c;
    // protect the first and last page, they are reserved for IP config and MAC address
    if (adr<512 || adr>7679)
        return;    
    // unlock the flash write protection
    cli();
    write_CP2200(FLASHKEY,0xA5);
    write_CP2200(FLASHKEY,0xF1);
    // erase the page
    write_CP2200(FLASHADDRL,adr & 0xFF);
    write_CP2200(FLASHADDRH,adr >> 8);
    write_CP2200(FLASHERASE,0x01);
    sei();
    // wait until erasing finishes
    u8_t flash_status;
    do {
        flash_status=read_CP2200(FLASHSTA);
    } while ((flash_status & 8)>0);
    do {
      c=string[0];
      writeflash(adr++,c);
      string++;
      // protect the last page, they are reserved for IP config and MAC address
      if (adr>7679)
        return;    
    }
    while (c!=0);
}


#ifdef EMAIL_APP

    #include <string.h>

    // Read a String from the flash memory at the given adress and encode it into HTML.
    void readFlashStringEncoded(char* buffer, u16_t adr) {
        char c;
        write_CP2200(FLASHADDRL,adr & 0xFF);
        write_CP2200(FLASHADDRH,adr >> 8);
        do {
            c=read_CP2200(FLASHAUTORD);
            // Replace ascii 255 by ascii 0.
            // Empty flash cells have that value
            if (c==255)
                c=0;
            // Replace forbidden characters by their escape sequences
            if (c=='&') {
                strcpy_P(buffer,PSTR("&amp;"));
                buffer+=5;
            }
            else if (c=='<') {
                strcpy_P(buffer,PSTR("&lt;"));
                buffer+=4;
            }
            else if (c=='>') {
                strcpy_P(buffer,PSTR("&gt;"));
                buffer+=4;
            }
            else if (c=='\"') {
                strcpy_P(buffer,PSTR("&quot;"));
                buffer+=6;
            }
            else {
                buffer[0]=c;
                buffer++;
            }
        }
        while (c!=0);
    }

#endif //EMAIL_APP
