#include <string.h>
#include <stdlib.h>
#include <avr/io.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include "httpd.h"
#include "httpd-fs.h"
#include "../uip/uip.h"
#include "../uip/uip_arp.h"
#include "../dhcpc/dhcpc.h"
#include "../driver/CP2200.h"
#include "../io-commands.h"
#include "../hw-layout.h"

#if ADC_CHANNELS > 0
    #include "../driver/ADC.h"
#endif

#ifdef EMAIL_APP
    #include "../email_app/email_app.h"
#endif 

#if defined SHIFTin || defined SHIFTout
   #include "../driver/shiftregisters.h"
#endif

#ifdef CARDSLOT
    // these variables are located in the main.c file.
    extern uint8_t card_present; // set to 1 during start-up if the SD card is accessible
    extern uint8_t card_busy;    // Used to ensure that only one single thread uses the avrfat32 library
#endif

// Functions for creating dynamic websites
//
// Functions can perform some action and the can optionally produce some 
// output directly in the ethernet buffer.
// The size of the result must fit into that buffer.
//
// Functions can be called in two ways.
// 1) Script marker in a shtml file: %! function_name?params
// 2) URL: http://hostname/function_name?parameters
//
// The function "io?params" calls functions from io-commands.c.
//
// To call a function via URL that does not use parameters, enter any
// dummy parameter, otherwise httpd thinks that this is a regular filename
// and no function call.
//
// To produce output, functions write a string into the buffer uip_appdata
// and return the size of the result. The result must not exceed the 
// buffers size, which is about 1400 bytes.
//
// If the name of the function is the same as a filename, this file gets
// loaded after the function returns.


// Show how often each file has been accessed

static unsigned short function_file_stats(char* buffer, int bufsize) {
    unsigned short len=0;
    // list as many files as fit into the buffer
    for (int i = 0; (len+30) < bufsize; i++) {
        // name fo the file
        char* name = httpd_fs_getName(i);
        // break if there is no file left anmyore
        if (name == 0) {
            break;
        }        
        // we cant simply pass the filename to snprintf because it is in
        // program memory and snprintf requires strings in RAM.
        int fnameLen=strlen_P(name);
        strcpy_P(buffer + len, name);        
        len += fnameLen;
        // append space character so that all filename have the same length
        for (int j=fnameLen; j<20; j++) {
            buffer[len++]=' ';
        }
        // append the file access counter
        int counter = httpd_fs_getCount(i);
        len += sprintf_P(buffer + len, PSTR("%5u\n"), counter);
    }
    return len;
}


//-----------------------------------------------------------------------

// parse an IP address

void parse_ip_address(u8_t result[], const char* source) {
    char c;
    int d0,d1,d2,d3; // needed because sscanf requires int variables
    sscanf_P(source,PSTR("%d%c%d%c%d%c%d"),&d0,&c,&d1,&c,&d2,&c,&d3);
    // copy the int values to the result bytes.
    result[0]=d0;
    result[1]=d1;
    result[2]=d2;
    result[3]=d3;
}

//-------------------------------------------------------------------------

// Show the largest available block of free memory

static unsigned short function_freememory(char* buffer) {
    //
    void *p = 0;
    size_t size;
    for (size = RAMEND; size > 1; size -= 10) {
        p = malloc(size);
        if (p != 0) {
            break;
        }
    }
    if (p != 0) {
        free(p);
    }
    else {
        size=0;
    }
    return sprintf_P(buffer,PSTR("%5u"),size);
}


//---------------------------------------------------------------------------

#if UIP_STATISTICS

const char ip_stat_formatter[] PROGMEM = 
    "%5u\n%5u\n%5u\n\n"
    "%5u\n%5u\n%5u\n%5u\n%5u\n%5u\n%5u\n\n"
    "%5u\n%5u\n%5u\n%5u\n\n"
    "%5u\n%5u\n%5u\n\n"
    "%5u\n%5u\n%5u\n%5u\n%5u\n%5u\n\n"
    "%5u\n%5u\n%5u\n%5u\n%5u";

// Show network I/O statistics from IP protocol

static unsigned short function_net_stats(char* buffer, const int bufsize) {
    return snprintf_P(buffer, bufsize, ip_stat_formatter,
            uip_stat.ip.recv,
            uip_stat.ip.sent,
            uip_stat.ip.drop,

            uip_stat.ip.vhlerr,
            uip_stat.ip.hblenerr,
            uip_stat.ip.lblenerr,
            uip_stat.ip.fragerr,
            uip_stat.ip.chkerr,
            uip_stat.ip.protoerr,
            uip_stat.ip.wronghost,

            uip_stat.icmp.recv,
            uip_stat.icmp.sent,
            uip_stat.icmp.drop,
            uip_stat.icmp.typeerr,

            uip_stat.tcp.recv,
            uip_stat.tcp.sent,
            uip_stat.tcp.drop,

            uip_stat.tcp.chkerr,
            uip_stat.tcp.ackerr,
            uip_stat.tcp.rst,
            uip_stat.tcp.rexmit,
            uip_stat.tcp.syndrop,
            uip_stat.tcp.synrst,
      #if UIP_CONF_UDP
            uip_stat.udp.recv,
            uip_stat.udp.sent,
            uip_stat.udp.drop,
            uip_stat.udp.chkerr,
            uip_stat.udp.wrongconn
      #else // UIP_CONF_UDP
            0,0,0,0,0
#endif // UIP_CONF_UDP
            );
}

#endif // UIP_STATISTIC



//---------------------------------------------------------------------------


const char tcp_closed[]      PROGMEM="closed";
const char tcp_syn_rcvd[]    PROGMEM="syn-rcvd";
const char tcp_syn_sent[]    PROGMEM="syn-sent";
const char tcp_established[] PROGMEM="established";
const char tcp_fin_wait_1[]  PROGMEM="fin-wait-1";
const char tcp_fin_wait_2[]  PROGMEM="fin-wait-2";
const char tcp_closing[]     PROGMEM="closing";
const char tcp_time_wait[]   PROGMEM="time-wait";
const char tcp_last_ack[]    PROGMEM="last-ack";

const char tcp_conn_formatter1[] PROGMEM = "<tr><td>%d</td><td>%u.%u.%u.%u:%u</td><td>";
const char tcp_conn_formatter2[] PROGMEM = "</td><td>%u</td><td>%u</td><td>%c %c</td></tr>\n";


// Show informations about all TCP connections

static unsigned short function_tcp_connections(char* buffer, const int bufsize) {
    unsigned short len = 0;
    // list as many connections as fit into the buffer
    for (int i = 0; len < bufsize && i<UIP_CONF_MAX_CONNECTIONS; i++) {
        len+=snprintf_P(buffer+len, bufsize-len, tcp_conn_formatter1,
                htons(uip_conns[i].lport),
                htons(uip_conns[i].ripaddr[0]) >> 8,
                htons(uip_conns[i].ripaddr[0]) & 0xff,
                htons(uip_conns[i].ripaddr[1]) >> 8,
                htons(uip_conns[i].ripaddr[1]) & 0xff,
                htons(uip_conns[i].rport));
        switch (uip_conns[i].tcpstateflags & UIP_TS_MASK) {
            case UIP_CLOSED:
                strncpy_P(buffer+len,tcp_closed,bufsize-len);
                len+=sizeof(tcp_closed)-1;
                break;
            case UIP_SYN_RCVD:
                strncpy_P(buffer+len,tcp_syn_rcvd,bufsize-len);
                len+=sizeof(tcp_syn_rcvd)-1;
                break;
            case UIP_SYN_SENT:
                strncpy_P(buffer+len,tcp_syn_sent,bufsize-len);
                len+=sizeof(tcp_syn_sent)-1;
                break;
            case UIP_ESTABLISHED:
                strncpy_P(buffer+len,tcp_established,bufsize-len);
                len+=sizeof(tcp_established)-1;
                break;
            case UIP_FIN_WAIT_1:
                strncpy_P(buffer+len,tcp_fin_wait_1,bufsize-len);
                len+=sizeof(tcp_fin_wait_1)-1;
                break;
            case UIP_FIN_WAIT_2:
                strncpy_P(buffer+len,tcp_fin_wait_2,bufsize-len);
                len+=sizeof(tcp_fin_wait_1)-1;
                break;
            case UIP_CLOSING:
                strncpy_P(buffer+len,tcp_closing,bufsize-len);
                len+=sizeof(tcp_closing);
                break;
            case UIP_TIME_WAIT:
                strncpy_P(buffer+len,tcp_time_wait,bufsize-len);
                len+=sizeof(tcp_time_wait)-1;
                break;
            case UIP_LAST_ACK:
                strncpy_P(buffer+len,tcp_last_ack,bufsize-len);
                len+=sizeof(tcp_last_ack)-1;
                break;
        }
        len+=snprintf_P(buffer+len, bufsize-len, tcp_conn_formatter2,
                uip_conns[i].nrtx,
                uip_conns[i].timer,
                (uip_outstanding(&uip_conns[i])) ? '*' : ' ',
                (uip_stopped(&uip_conns[i])) ? '!' : ' ');
    }
    return len;
}


//---------------------------------------------------------------------------

// get ip address, gateways, netmask  and authentication 

const char tcp_address_formatter[] PROGMEM = "%u.%u.%u.%u";

// get the current ip address
static unsigned short function_get_current_ip(char* buffer) {
    u8_t adr[4];
    uip_gethostaddr(adr);
    return sprintf_P(buffer, tcp_address_formatter,adr[0],adr[1],adr[2],adr[3]);
}

// get the current netmask
static unsigned short function_get_current_mask(char* buffer) {
    u8_t adr[4];
    uip_getnetmask(adr);
    return sprintf_P(buffer, tcp_address_formatter,adr[0],adr[1],adr[2],adr[3]);
}

// get the current gateway
static unsigned short function_get_current_gw(char* buffer) {
    u8_t adr[4];
    uip_getdraddr(adr);
    return sprintf_P(buffer, tcp_address_formatter,adr[0],adr[1],adr[2],adr[3]);
}

// get the configured ip address
static unsigned short function_get_configured_ip(char* buffer) {
    return sprintf_P(buffer, tcp_address_formatter,config.ipaddr[0],config.ipaddr[1],config.ipaddr[2],config.ipaddr[3]);
}

// get the configured netmask
static unsigned short function_get_configured_mask(char* buffer) {
    return sprintf_P(buffer, tcp_address_formatter,config.netmask[0],config.netmask[1],config.netmask[2],config.netmask[3]);
}

// get the configured gateway
static unsigned short function_get_configured_gw(char* buffer) {
    return sprintf_P(buffer,  tcp_address_formatter,config.gateway[0],config.gateway[1],config.gateway[2],config.gateway[3]);
}

// get the configured authentication username:password
static unsigned short function_get_configured_auth(char* buffer) {
    if (config.authentication[0]==0 || config.authentication[0]==255) {
        return 0;
    }
    else {
        strcpy(buffer,config.authentication);
        return strlen(config.authentication);
    }
}



//------------------------------------------------------------------------------

// get a dhcp enabled flag for the config form

static unsigned short function_dhcp_selected(char* buffer) {
    if (config.dhcpenable) {
        return sprintf_P(buffer, PSTR("selected"));
    }
    else {
        // return a dummy because a null string is not allowed
        buffer[0]=' ';
        return 1;
    }
}

//------------------------------------------------------------------------------

// get MAC address

static unsigned short function_get_mac(char* buffer) {
    return sprintf_P(buffer, PSTR("%02x:%02x:%02x:%02x:%02x:%02x"),
            uip_ethaddr.addr[0],
            uip_ethaddr.addr[1],
            uip_ethaddr.addr[2],
            uip_ethaddr.addr[3],
            uip_ethaddr.addr[4],
            uip_ethaddr.addr[5]);
}


//------------------------------------------------------------------------------

const char dhcp[] PROGMEM = "dhcp=";
const char ip[]   PROGMEM = "ip=";
const char mask[] PROGMEM = "mask=";
const char gw[]   PROGMEM = "gw=";
const char auth[] PROGMEM = "auth=";

// process form data from config.shtml

static unsigned short function_config(char* buffer, const char* parameters) {
    // find the settings in parameters and parse them
    char* position;
    int value;
    position=strstr_P(parameters,dhcp);
    if (position!=NULL) {
        value=atoi(position+sizeof(dhcp)-1);
        config.dhcpenable=value;
    }
    position=strstr_P(parameters,ip);
    if (position!=NULL) {
        parse_ip_address(config.ipaddr,position+sizeof(ip)-1);
        uip_sethostaddr(&config.ipaddr);
    }
    position=strstr_P(parameters,mask);
    if (position!=NULL) {
        parse_ip_address(config.netmask,position+sizeof(mask)-1);
        uip_setnetmask(&config.netmask);
    }
    position=strstr_P(parameters,gw);
    if (position!=NULL) {
        parse_ip_address(config.gateway,position+sizeof(gw)-1);
        uip_setdraddr(&config.gateway);
    }
    position=strstr_P(parameters,auth);
    if (position!=NULL) {
        char* param=position+sizeof(auth)-1;
        url_decode(param);
        // copy the parameter to the config
        strcpy(config.authentication,param);
    }
    writeFlashConfig();
    // return a dummy because a null string is not allowed
    buffer[0]=' ';
    buffer[1]=0;
    return 1;
}

//--------------------------------------------------------------------------

// Encode a number to string in binary format
// width is the number of bits to use (1-32)

void encodeBinary(char* buffer, const uint32_t value, const uint8_t width) {
    uint32_t val=value;
    int posi=width-1;        
    for (int8_t bit=0; bit<width; bit++) {
        if (val & 1) {
            buffer[posi]='1';
        }
        else {
            buffer[posi]='0';
        }
        posi--;
        val=(uint32_t) val>>1;
    }        
    // terminate the result string
    buffer[width]=0;
}

// Print status of a one regular AVR port
// returns the number of character that have been written into the buffer

int print_one_port(char* buffer, char port, uint8_t value) {
    unsigned short len=sprintf_P(buffer, PSTR("Port %c: 0x%02X = "), port, value);
    encodeBinary(buffer+len,value,8); len+=8;    
    buffer[len++]='\n';
    return len;
}

// Print status of a all regular AVR ports.

static unsigned short function_port_status(char* buffer) {   
    unsigned short len=0;
    
    // For ATmega
    #ifdef PINA
      len+=print_one_port(buffer+len,'A',PINA);
    #endif
    #ifdef PINB
      len+=print_one_port(buffer+len,'B',PINB);
    #endif
    #ifdef PINC
      len+=print_one_port(buffer+len,'C',PINC);
    #endif
    #ifdef PIND
      len+=print_one_port(buffer+len,'D',PIND);
    #endif
    #ifdef PINE
      len+=print_one_port(buffer+len,'E',PINE);
    #endif
    #ifdef PINF
      len+=print_one_port(buffer+len,'F',PINF);
    #endif
    #ifdef PING
      len+=print_one_port(buffer+len,'G',PING);
    #endif
    #ifdef PINH
      len+=print_one_port(buffer+len,'H',PINH);
    #endif
      
    // For Xmega
    #ifdef PORTA_IN
      len+=print_one_port(buffer+len,'A',PORTA_IN);
    #endif
    #ifdef PORTB_IN
      len+=print_one_port(buffer+len,'B',PORTB_IN);
    #endif
    #ifdef PORTC_IN
      len+=print_one_port(buffer+len,'C',PORTC_IN);
    #endif
    #ifdef PORTD_IN
      len+=print_one_port(buffer+len,'D',PORTD_IN);
    #endif
    #ifdef PORTE_IN
      len+=print_one_port(buffer+len,'E',PORTE_IN);
    #endif
    #ifdef PORTF_IN
      len+=print_one_port(buffer+len,'F',PORTF_IN);
    #endif
    #ifdef PORTG_IN
      len+=print_one_port(buffer+len,'G',PORTG_IN);
    #endif
    #ifdef PORTH_IN
      len+=print_one_port(buffer+len,'H',PORTH_IN);
    #endif      
    return len;
}


// Print status of the extension ports.

static unsigned short function_port_XL_status(char* buffer) {   
    unsigned short len=0;
      
    #ifdef SHIFTin
    {
        uint32_t value32=read_XL_inputs();
        #if XL_INPUTS > 24
          len+=sprintf_P(buffer+len, PSTR("Port X: 0x%08lX = "), value32);
        #elif XL_INPUTS > 16
          len+=sprintf_P(buffer+len, PSTR("Port X: 0x%06lX = "), value32);
        #elif XL_INPUTS > 8
          len+=sprintf_P(buffer+len, PSTR("Port X: 0x%04lX = "), value32);
        #else
          len+=sprintf_P(buffer+len, PSTR("Port X: 0x%02lX = "), value32);
        #endif 
        encodeBinary(buffer+len, value32, XL_INPUTS); 
        len+=XL_INPUTS;
        buffer[len++]='\n';
    }
    #endif 
    
    #ifdef SHIFTout
    {
        uint32_t value32=xlOutputState;
        #if XL_OUTPUTS > 24
          len+=sprintf_P(buffer+len, PSTR("Port Y: 0x%08lX = "), value32);
        #elif XL_OUTPUTS > 16
          len+=sprintf_P(buffer+len, PSTR("Port Y: 0x%06lX = "), value32);
        #elif XL_OUTPUTS > 8
          len+=sprintf_P(buffer+len, PSTR("Port Y: 0x%04lX = "), value32);
        #else
          len+=sprintf_P(buffer+len, PSTR("Port Y: 0x%02lX = "), value32);
        #endif
        encodeBinary(buffer+len, value32, XL_OUTPUTS); 
        len+=XL_OUTPUTS;
        buffer[len++]='\n';
    }
    #endif
    return len;
}


//--------------------------------------------------------------------------


// Status of all ADC channels.

static unsigned short function_adc_status(char* buffer) {
    unsigned short len=0;
    #if ADC_CHANNELS > 0
        for (int i=0; i<ADC_CHANNELS; i++) {
            int value=getADC(i);
            len+=sprintf_P(buffer+len, PSTR("ADC %i: 0x0%03X\n"),i,value);
        }
    #else
      buffer[0]=0; // empty string
    #endif
    return len;
}


//--------------------------------------------------------------------------

// Status of the SD/MMC card slot

static unsigned short function_mmcStatus(char* buffer) {
    #ifdef CARDSLOT
        if (card_present && card_busy) {
            strcpy_P(buffer,PSTR("yes, busy"));
            return 9;
        }
        else  if (card_present) {
            strcpy_P(buffer,PSTR("yes, idle"));
            return 9;
        }
        else {
            strcpy_P(buffer,PSTR("no"));
            return 2;
        }
    #else
        strcpy_P(buffer,PSTR("n.a"));
        return 4;
    #endif
}

//--------------------------------------------------------------------------

// Call an I/O command from io-commands.c

static unsigned short function_io(char* buffer, const char* parameters) {    
    // execute the io command
    io_command(buffer,parameters);
    return strlen(buffer);        
}


//------------------------------------------------------------------------------

const char hiddenstyle[]            PROGMEM = "display:none";

// Calls a function by name
// state->filename is the function name string
// state->parameters is the parameters string

unsigned short httpd_function_dispatcher(void* state) {
    struct application_state* app_state=state;
    char* name = app_state->filename;
    char* parameters = app_state->parameters;
    char* buffer = uip_appdata;
    int bufsize=uip_mss();
    
    if (strcmp_P(name, PSTR("io"))==0) {
        return function_io(buffer,parameters);
    }
    else if (strcmp_P(name, PSTR("file-stats")) == 0) {
        return function_file_stats(buffer,bufsize);
    }
    #if UIP_STATISTICS
        else if (strcmp_P(name, PSTR("net-stats")) == 0) {
            return function_net_stats(buffer,bufsize);
        }
    #endif // UIP_STATISTICS
    else if (strcmp_P(name, PSTR("tcp-conn")) == 0) {
        return function_tcp_connections(buffer,bufsize);
    }
    else if (strcmp_P(name, PSTR("freemem")) == 0) {
        return function_freememory(buffer);
    }
    #if UIP_CONF_UDP
        else if (strcmp_P(name, PSTR("dhcp-status")) == 0) {
            return get_dhcp_status(buffer);
        }
        else if (strcmp_P(name, PSTR("dhcpsel")) == 0) {
            return function_dhcp_selected(buffer);
        }
    #endif // UIP_CONF_UDP
    else if (strcmp_P(name, PSTR("getip")) == 0) {
        return function_get_configured_ip(buffer);
    }
    else if (strcmp_P(name, PSTR("getmask")) == 0) {
        return function_get_configured_mask(buffer);
    }
    else if (strcmp_P(name, PSTR("getgw")) == 0) {
        return function_get_configured_gw(buffer);
    }
    else if (strcmp_P(name, PSTR("getauth")) == 0) {
        return function_get_configured_auth(buffer);
    }
    else if (strcmp_P(name, PSTR("currentip")) == 0) {
        return function_get_current_ip(buffer);
    }
    else if (strcmp_P(name, PSTR("currentmask")) == 0) {
        return function_get_current_mask(buffer);
    }
    else if (strcmp_P(name, PSTR("currentgw")) == 0) {
        return function_get_current_gw(buffer);
    }
    else if (strcmp_P(name, PSTR("config.shtml")) == 0) {
        return function_config(buffer, parameters);
    }
    else if (strcmp_P(name, PSTR("getmac")) == 0) {
        return function_get_mac(buffer);
    }
    else if (strcmp_P(name,PSTR("portstatus"))==0) {
        return function_port_status(buffer);
    }
    else if (strcmp_P(name,PSTR("portXLstatus"))==0) {
        return function_port_XL_status(buffer);
    }
    else if (strcmp_P(name,PSTR("adcstatus"))==0) {
        return function_adc_status(buffer);
    }
    else if (strcmp_P(name,PSTR("mmcStatus"))==0) {
        return function_mmcStatus(buffer);
    }    
    else if (strcmp_P(name,PSTR("hasMAIL"))==0) {
        #ifdef EMAIL_APP
            return 0;
        #else
            strcpy_P(buffer,hiddenstyle);
            return sizeof(hiddenstyle)-1;
        #endif
    }        
    else if (strcmp_P(name,PSTR("noMAIL"))==0) {
        #ifdef EMAIL_APP
            strcpy_P(buffer,hiddenstyle);
            return sizeof(hiddenstyle)-1;
        #else
            return 0;
        #endif
    }   
    else if (strcmp_P(name,PSTR("canauth"))==0) {
        #ifdef AUTH
            return 0;
        #else
            strcpy_P(buffer,hiddenstyle);
            return sizeof(hiddenstyle)-1;
        #endif 
    }
    #ifdef EMAIL_APP 
        else {
            int i=email_httpd_functions(buffer,bufsize,name,parameters);
            if (i>=0) {
                return i;
            }
        }
    #endif 
    
    // the function name is unknown.
    return sprintf_P(buffer,PSTR("Unknown function %s, params=%s\n"),name,parameters);
}

