#include "io-commands.h"
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "hw-layout.h"

// Universal commands to control the external I/O lines.

#if ADC_CHANNELS > 0
   #include "driver/ADC.h"
#endif

#if defined SHIFTin || defined SHIFTout
   #include "driver/shiftregisters.h"
#endif


#if defined DDRA || defined DDRB || defined DDRC || defined DDRD 
    // Macros for ATmega

    // Configure pull-Ups on all unprotected input pins of a port
    #define pullup_port(port,ddr,protect,value) { \
        port=(value & ~(ddr | protect)) | (port & (ddr | protect)); \
    }
    
    // Configure pull-up of a single input pin if it is not protected
    #define pullup_pin(port,ddr,protect,pin,value) { \
        if ((protect & (1<<pin))==0 && readBit(ddr,pin)==0)  \
            writeBit(port,pin,value); \
    }
            
    // Set direction of all unprotected pins of a port   
    #define dir_port(port,ddr,protect,value) { \
        ddr=(value & ~protect) | (ddr & protect); \
    }
    
    // Set direction of a single port pin if it is unprotected   
    #define dir_pin(port,ddr,protect,pin,value) { \
        if ((protect & (1<<pin))==0) \
            writeBit(ddr,pin,value); \
    }
    
    // Write to all unprotected output pins of a port
    #define out_port(port,ddr,protect,value) { \
        port=(value & (ddr & ~protect)) | (port & (~ddr | protect)); \
    }
    
    // Write to a single unprotected output pin of a port
    #define out_pin(port,ddr,protect,pin,value) { \
        if ((readBit(ddr,pin)==1) && ((protect & (1<<pin))==0)) \
            writeBit(port,pin,value); \
    }
    
    // Read from an input port
    #define in_port(port,port_in) (port_in) 
    
    // Read from an input pin
    #define in_pin(port,port_in,pin) readBit(port_in,pin)

#else
    // Macros for Xmega
    
    // Configure pull-Ups on all unprotected input pins of a ports
    #define pullup_port(port,ddr,protect,value) { \
        PORTCFG_MPCMASK = (uint8_t) (value & ~protect); \
        port.PIN0CTRL = PORT_OPC_PULLUP_gc; \
        PORTCFG_MPCMASK = (uint8_t) (~value & ~protect); \
        port.PIN0CTRL = 0; \
    }
    
    // Configure pull-up of a single input pin if it is not protected
    #define pullup_pin(port,ddr,protect,pin,value) { \
        if ((protect & (1<<pin))==0 && readBit(port.DIR,pin)==0) { \
            PORTCFG_MPCMASK = (uint8_t) ((1<<pin) & ~protect); \
            if (value) \
                port.PIN0CTRL = PORT_OPC_PULLUP_gc; \
            else \
                port.PIN0CTRL = 0; \
        } \
    }

    // Set direction of all unprotected pins of a port   
    #define dir_port(port,ddr,protect,value) { \
        port.DIR=(value & ~protect) | (port.DIR & protect); \
    }
    
    // Set direction of a single port pin if it is unprotected   
    #define dir_pin(port,ddr,protect,pin,value) { \
        if ((protect & (1<<pin))==0) \
            writeBit(port.DIR,pin,value); \
    }
    
    // Write to all unprotected output pins of a port
    #define out_port(port,ddr,protect,value) { \
        port.OUT=(value & (port.DIR & ~protect)) | (port.OUT & (~port.DIR | protect)); \
    }
    
    // Write to a single unprotected output pin of a port
    #define out_pin(port,ddr,protect,pin,value) { \
        if ((readBit(port.DIR,pin)==1) && ((protect & (1<<pin))==0)) \
            writeBit(port.OUT,pin,value); \
    }
    
    // Read from an input port
    #define in_port(port,port_in) (port.IN) 
    
    // Read from an input pin
    #define in_pin(port,port_in,pin) readBit(port.IN,pin)


#endif

const char ok[]                 PROGMEM="Ok\n";
const char bad_command[]        PROGMEM="Bad command\n";
const char invalid_port[]       PROGMEM="Invalid port\n";
const char invalid_pin[]        PROGMEM="Invalid pin\n";
const char invalid_hex[]        PROGMEM="Invalid hex number\n";
const char invalid_boolean[]    PROGMEM="Invalid boolean value\n";
const char invalid_channel[]    PROGMEM="Invalid channel number\n";
const char invalid_reference[]  PROGMEM="Invalid reference name\n";

// Decode a hexadecimal digit to integer
// Invalid characters return -1.

int16_t decodeHex1(const char c) {
    if (c>='0' && c<='9')
        return c-'0';
    else if (c>='A' && c<='F')
        return c-'A'+10;
    else if (c>='a' && c<='f')
        return c-'a'+10;
    else
        return -1;
}

// Decode a 2-digit hexadecimal string to integer
// Note that the string length is not checked.
// Invalid characters return -1.

int16_t decodeHex(const char* characters) {
    int16_t highNibble=decodeHex1(characters[0]);
    int16_t lowNibble=decodeHex1(characters[1]);
    if ((highNibble<0) || (lowNibble<0))
        return -1;
    else
        return ((highNibble<<4) | (lowNibble));
}


// Decode a single boolean character to integer
// Invalid characters return -1.

int16_t decodeBin(const char c) {
    switch (c) {
        case '1': return 1;
        case '0': return 0;
        case 'H': return 1;
        case 'L': return 0;
        case 'E': return 1;
        case 'D': return 0;
        case 'O': return 1;
        case 'I': return 0;
        case 'h': return 1;
        case 'l': return 0;
        case 'e': return 1;
        case 'd': return 0;
        case 'o': return 1;
        case 'i': return 0;
        default: return -1;
    }
}

// Decode a decimal number to integer.
// Invalid characters or empty string return -1.

int16_t decodeDec(const char* s) {
    if (s[0]<'0' || s[0]>'9')
        return -1;
    else
        return atoi(s);
}


//-----------------------------------------------------------------------------------------------

// Execute an p command
//     p=enable pull-up's on input pins

void command_p(char* result, const char* command) {
    size_t len=strlen(command);
    if (command[1]=='=' || command[1]==',') {
        if (len>10) { // write to 8 ports at once
            int16_t valueH=decodeHex(command+2);
            int16_t valueG=decodeHex(command+4);
            int16_t valueF=decodeHex(command+6);
            int16_t valueE=decodeHex(command+8);
            int16_t valueD=decodeHex(command+10);
            int16_t valueC=decodeHex(command+12);
            int16_t valueB=decodeHex(command+14);
            int16_t valueA=decodeHex(command+16);
            if (valueH<0 || valueG<0 || valueF<0 || valueE<0 ||valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                pullup_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                pullup_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                pullup_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                pullup_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            #ifdef PORTE
                pullup_port(PORTE,DDRE,protect_PORTE,valueE);
            #endif
            #ifdef PORTF
                pullup_port(PORTF,DDRF,protect_PORTF,valueF);
            #endif
            #ifdef PORTG
                pullup_port(PORTG,DDRG,protect_PORTG,valueG);
            #endif
            #ifdef PORTH
                pullup_port(PORTH,DDRH,protect_PORTH,valueH);
            #endif
            strcpy_P(result,ok);
        }
        else  { // write to 4 ports at once
            int16_t valueD=decodeHex(command+2);
            int16_t valueC=decodeHex(command+4);
            int16_t valueB=decodeHex(command+6);
            int16_t valueA=decodeHex(command+8);
            if (valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                pullup_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                pullup_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                pullup_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                pullup_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            strcpy_P(result,ok);
        }
    }
    else if (command[3]=='=' || command[3]==',') { // write to one port (8 bits)
         if (command[1]!='P') {
             strcpy_P(result,invalid_port);
             return;
         }
         char port=command[2];
        int value=decodeHex(command+4);
        if (value<0) {
            strcpy_P(result,invalid_hex);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': pullup_port(PORTA,DDRA,protect_PORTA,value); break;
            #endif
            #ifdef PORTB
                case 'B': pullup_port(PORTB,DDRB,protect_PORTB,value); break;
            #endif
            #ifdef PORTC
                case 'C': pullup_port(PORTC,DDRC,protect_PORTC,value); break;
            #endif
            #ifdef PORTD
                case 'D': pullup_port(PORTD,DDRD,protect_PORTD,value); break;
            #endif
            #ifdef PORTE
                case 'E': pullup_port(PORTE,DDRE,protect_PORTE,value); break;
            #endif
            #ifdef PORTF
                case 'F': pullup_port(PORTF,DDRF,protect_PORTF,value); break;
            #endif
            #ifdef PORTG
                case 'G': pullup_port(PORTG,DDRG,protect_PORTG,value); break;
            #endif
            #ifdef PORTH
                case 'H': pullup_port(PORTH,DDRH,protect_PORTH,value); break;
            #endif
            #ifdef PORTI
                case 'I': pullup_port(PORTI,DDRI,protect_PORTI,value); break;
            #endif
            #ifdef PORTJ
                case 'J': pullup_port(PORTJ,DDRJ,protect_PORTJ,value); break;
            #endif
            #ifdef PORTK
                case 'K': pullup_port(PORTK,DDRK,protect_PORTK,value); break;
            #endif
            #ifdef PORTL
                case 'L': pullup_port(PORTL,DDRL,protect_PORTL,value); break;
            #endif
             default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
    else if (command[4]=='=' || command[4]==',') { // write to a single port pin
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int pin=decodeDec(command+3);
        int value=decodeBin(command[5]);
        if (pin<0 || pin>7) {
            strcpy_P(result,invalid_pin);
            return;
        }
        else if (value<0 || value>1) {
            strcpy_P(result,invalid_boolean);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': pullup_pin(PORTA,DDRA,protect_PORTA,pin,value); break;
            #endif
            #ifdef PORTB
                case 'B': pullup_pin(PORTB,DDRB,protect_PORTB,pin,value); break;
            #endif
            #ifdef PORTC
                case 'C': pullup_pin(PORTC,DDRC,protect_PORTC,pin,value); break;
            #endif
            #ifdef PORTD
                case 'D': pullup_pin(PORTD,DDRD,protect_PORTD,pin,value); break;
            #endif
            #ifdef PORTE
                case 'E': pullup_pin(PORTE,DDRE,protect_PORTE,pin,value); break;
            #endif
            #ifdef PORTF
                case 'F': pullup_pin(PORTF,DDRF,protect_PORTF,pin,value); break;
            #endif
            #ifdef PORTG
                case 'G': pullup_pin(PORTG,DDRG,protect_PORTG,pin,value); break;
            #endif
            #ifdef PORTH
                case 'H': pullup_pin(PORTH,DDRH,protect_PORTH,pin,value); break;
            #endif
            #ifdef PORTI
                case 'I': pullup_pin(PORTI,DDRI,protect_PORTI,pin,value); break;
            #endif
            #ifdef PORTJ
                case 'J': pullup_pin(PORTJ,DDRJ,protect_PORTJ,pin,value); break;
            #endif
            #ifdef PORTK
                case 'K': pullup_pin(PORTK,DDRK,protect_PORTK,pin,value); break;
            #endif
            #ifdef PORTL
                case 'L': pullup_pin(PORTL,DDRL,protect_PORTL,pin,value); break;
            #endif
            default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
    else
        strcpy_P(result,bad_command);
}



//-----------------------------------------------------------------------------------------------

// Execute an d command
//     d=set i/o direction
//     value 1=output
//     value 0=input

void command_d(char* result, const char* command) {
    size_t len=strlen(command);
    if (command[1]=='=' || command[1]==',') {
        if (len>10) {  // write to 8 ports at once
            int valueH=decodeHex(command+2);
            int valueG=decodeHex(command+4);
            int valueF=decodeHex(command+6);
            int valueE=decodeHex(command+8);
            int valueD=decodeHex(command+10);
            int valueC=decodeHex(command+12);
            int valueB=decodeHex(command+14);
            int valueA=decodeHex(command+16);
            if (valueH<0 || valueG<0 || valueF<0 || valueE<0 ||valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                dir_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                dir_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                dir_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                dir_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            #ifdef PORTE
                dir_port(PORTE,DDRE,protect_PORTE,valueE);
            #endif
            #ifdef PORTF
                dir_port(PORTF,DDRF,protect_PORTF,valueF);
            #endif
            #ifdef PORTG
                dir_port(PORTG,DDRG,protect_PORTG,valueG);
            #endif
            #ifdef PORTH
                dir_port(PORTH,DDRH,protect_PORTH,valueH);
            #endif
            strcpy_P(result,ok);
        }
        else { // write to 4 ports at once
            int valueD=decodeHex(command+2);
            int valueC=decodeHex(command+4);
            int valueB=decodeHex(command+6);
            int valueA=decodeHex(command+8);
            if (valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                dir_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                dir_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                dir_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                dir_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            strcpy_P(result,ok);
        }
    }
    else if (command[3]=='=' || command[3]==',') { // write to a single port
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int value=decodeHex(command+4);
        if (value<0) {
            strcpy_P(result,invalid_hex);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': dir_port(PORTA,DDRA,protect_PORTA,value); break;
            #endif
            #ifdef PORTB
                case 'B': dir_port(PORTB,DDRB,protect_PORTB,value); break;
            #endif
            #ifdef PORTC
                case 'C': dir_port(PORTC,DDRC,protect_PORTC,value); break;
            #endif
            #ifdef PORTD
                case 'D': dir_port(PORTD,DDRD,protect_PORTD,value); break;
            #endif
            #ifdef PORTE
                case 'E': dir_port(PORTE,DDRE,protect_PORTE,value); break;
            #endif
            #ifdef PORTF
                case 'F': dir_port(PORTF,DDRF,protect_PORTF,value); break;
            #endif
            #ifdef PORTG
                case 'G': dir_port(PORTG,DDRG,protect_PORTG,value); break;
            #endif
            #ifdef PORTH
                case 'H': dir_port(PORTH,DDRH,protect_PORTH,value); break;
            #endif
            #ifdef PORTI
                case 'I': dir_port(PORTI,DDRI,protect_PORTI,value); break;
            #endif
            #ifdef PORTJ
                case 'J': dir_port(PORTJ,DDRJ,protect_PORTJ,value); break;
            #endif
            #ifdef PORTK
                case 'K': dir_port(PORTK,DDRK,protect_PORTK,value); break;
            #endif
            #ifdef PORTL
                case 'L': dir_port(PORTL,DDRL,protect_PORTL,value); break;
            #endif
            default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
    else if (command[4]=='=' || command[4]==',') { // write to a single port pin
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int pin=decodeDec(command+3);
        int value=decodeBin(command[5]);
        if (pin<0 || pin>7) {
            strcpy_P(result,invalid_pin);
            return;
        }
        else if (value<0 || value>1) {
            strcpy_P(result,invalid_boolean);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': dir_pin(PORTA,DDRA,protect_PORTA,pin,value); break;
            #endif
            #ifdef PORTB
                case 'B': dir_pin(PORTB,DDRB,protect_PORTB,pin,value); break;
            #endif
            #ifdef PORTC
                case 'C': dir_pin(PORTC,DDRC,protect_PORTC,pin,value); break;
            #endif
            #ifdef PORTD
                case 'D': dir_pin(PORTD,DDRD,protect_PORTD,pin,value); break;
            #endif
            #ifdef PORTE
                case 'E': dir_pin(PORTE,DDRE,protect_PORTE,pin,value); break;
            #endif
            #ifdef PORTF
                case 'F': dir_pin(PORTF,DDRF,protect_PORTF,pin,value); break;
            #endif
            #ifdef PORTG
                case 'G': dir_pin(PORTG,DDRG,protect_PORTG,pin,value); break;
            #endif
            #ifdef PORTH
                case 'H': dir_pin(PORTH,DDRH,protect_PORTH,pin,value); break;
            #endif
            #ifdef PORTI
                case 'I': dir_pin(PORTI,DDRI,protect_PORTI,pin,value); break;
            #endif
            #ifdef PORTJ
                case 'J': dir_pin(PORTJ,DDRJ,protect_PORTJ,pin,value); break;
            #endif
            #ifdef PORTK
                case 'K': dir_pin(PORTK,DDRK,protect_PORTK,pin,value); break;
            #endif
            #ifdef PORTL
                case 'L': dir_pin(PORTL,DDRL,protect_PORTL,pin,value); break;
            #endif
            default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
    else
        strcpy_P(result,bad_command);
}

//-----------------------------------------------------------------------------------------------

// Execute an o command
//     o=output data

void command_o(char* result, const char* command) {
    size_t len=strlen(command);
    if (command[1]=='=' || command[1]==',') {
        if (len>10) { // write to 8 ports at once
            int valueH=decodeHex(command+2);
            int valueG=decodeHex(command+4);
            int valueF=decodeHex(command+6);
            int valueE=decodeHex(command+8);
            int valueD=decodeHex(command+10);
            int valueC=decodeHex(command+12);
            int valueB=decodeHex(command+14);
            int valueA=decodeHex(command+16);
            if (valueH<0 || valueG<0 || valueF<0 || valueE<0 ||valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                out_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                out_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                out_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                out_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            #ifdef PORTE
                out_port(PORTE,DDRE,protect_PORTE,valueE);
            #endif
            #ifdef PORTF
                out_port(PORTF,DDRF,protect_PORTF,valueF);
            #endif
            #ifdef PORTG
                out_port(PORTG,DDRG,protect_PORTG,valueG);
            #endif
            #ifdef PORTH
                out_port(PORTH,DDRH,protect_PORTH,valueH);
            #endif
            strcpy_P(result,ok);
        }
        else { // write to 4 ports at once
            int valueD=decodeHex(command+2);
            int valueC=decodeHex(command+4);
            int valueB=decodeHex(command+6);
            int valueA=decodeHex(command+8);
            if (valueD<0 || valueC<0 || valueB<0 || valueA<0) {
                strcpy_P(result,invalid_hex);
                return;
            }
            #ifdef PORTA
                out_port(PORTA,DDRA,protect_PORTA,valueA);
            #endif
            #ifdef PORTB
                out_port(PORTB,DDRB,protect_PORTB,valueB);
            #endif
            #ifdef PORTC
                out_port(PORTC,DDRC,protect_PORTC,valueC);
            #endif
            #ifdef PORTD
                out_port(PORTD,DDRD,protect_PORTD,valueD);
            #endif
            strcpy_P(result,ok);
        }
    }
    else if (command[3]=='=' || command[3]==',') { // write to a single port
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int value=decodeHex(command+4);
        if (value<0) {
            strcpy_P(result,invalid_hex);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': out_port(PORTA,DDRA,protect_PORTA,value); break;
            #endif
            #ifdef PORTB
                case 'B': out_port(PORTB,DDRB,protect_PORTB,value); break;
            #endif
            #ifdef PORTC
                case 'C': out_port(PORTC,DDRC,protect_PORTC,value); break;
            #endif
            #ifdef PORTD
                case 'D': out_port(PORTD,DDRD,protect_PORTD,value); break;
            #endif
            #ifdef PORTE
                case 'E': out_port(PORTE,DDRE,protect_PORTE,value); break;
            #endif
            #ifdef PORTF
                case 'F': out_port(PORTF,DDRF,protect_PORTF,value); break;
            #endif
            #ifdef PORTG
                case 'G': out_port(PORTG,DDRG,protect_PORTG,value); break;
            #endif
            #ifdef PORTH
                case 'H': out_port(PORTH,DDRH,protect_PORTH,value); break;
            #endif
            #ifdef PORTI
                case 'I': out_port(PORTI,DDRI,protect_PORTI,value); break;
            #endif
            #ifdef PORTJ
                case 'J': out_port(PORTJ,DDRJ,protect_PORTJ,value); break;
            #endif
            #ifdef PORTK
                case 'K': out_port(PORTK,DDRK,protect_PORTK,value); break;
            #endif
            #ifdef PORTL
                case 'L': out_port(PORTL,DDRL,protect_PORTL,value); break;
            #endif

            #ifdef SHIFTout
                case 'Y': {
                        // value is 8 bit, parse up to 32bit in total
                        uint32_t value32=value;
                        if (command[6]!=0) {
                            value=decodeHex(command+6);
                            if (value<0) {
                                strcpy_P(result,invalid_hex);
                                return;
                            }
                            value32=(uint32_t) (value32 << 8) | value;
                            if (command[8]!=0) {
                                value=decodeHex(command+8);
                                if (value<0) {
                                    strcpy_P(result,invalid_hex);
                                    return;
                                }
                                value32=(uint32_t)(value32 << 8) | value;
                                if (command[10]!=0) {
                                    value=decodeHex(command+10);
                                    if (value<0) {
                                        strcpy_P(result,invalid_hex);
                                        return;
                                    }
                                    value32=(uint32_t)(value32 << 8) | value;
                                }                                         
                            }
                        }
                        write_XL_outputs(value32);
                    }                            
                    break;
            #endif
                    
            default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
    else { // write to a single port pin
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int pin=decodeDec(command+3);
        #ifdef SHIFTin
            if (port=='X') {
                if (pin<0 || pin>=XL_INPUTS) {
                    strcpy_P(result,invalid_pin);
                    return;
                }            
            }
            else 
        #endif
        #ifdef SHIFTout
            if (port=='Y') {
                if (pin<0 || pin>=XL_OUTPUTS) {
                    strcpy_P(result,invalid_pin);
                    return;
                }            
            }
            else 
        #endif
        if (pin<0 || pin>7) {
            strcpy_P(result,invalid_pin);
            return;
        }
        int value;
        if (pin>9) {
            value=decodeBin(command[6]);
        }
        else {
            value=decodeBin(command[5]);
        }
        if (value<0 || value>1) {
            strcpy_P(result,invalid_boolean);
            return;
        }
        switch (port) {
            #ifdef PORTA
                case 'A': out_pin(PORTA,DDRA,protect_PORTA,pin,value); break;
            #endif
            #ifdef PORTB
                case 'B': out_pin(PORTB,DDRB,protect_PORTB,pin,value); break;
            #endif
            #ifdef PORTC
                case 'C': out_pin(PORTC,DDRC,protect_PORTC,pin,value); break;
            #endif
            #ifdef PORTD
                case 'D': out_pin(PORTD,DDRD,protect_PORTD,pin,value); break;
            #endif
            #ifdef PORTE
                case 'E': out_pin(PORTE,DDRE,protect_PORTE,pin,value); break;
            #endif
            #ifdef PORTF
                case 'F': out_pin(PORTF,DDRF,protect_PORTF,pin,value); break;
            #endif
            #ifdef PORTG
                case 'G': out_pin(PORTG,DDRG,protect_PORTG,pin,value); break;
            #endif
            #ifdef PORTH
                case 'H': out_pin(PORTH,DDRH,protect_PORTH,pin,value); break;
            #endif
            #ifdef PORTI
                case 'I': out_pin(PORTI,DDRI,protect_PORTI,pin,value); break;
            #endif
            #ifdef PORTJ
                case 'J': out_pin(PORTJ,DDRJ,protect_PORTJ,pin,value); break;
            #endif
            #ifdef PORTK
                case 'K': out_pin(PORTK,DDRK,protect_PORTK,pin,value); break;
            #endif
            #ifdef PORTL
                case 'L': out_pin(PORTL,DDRL,protect_PORTL,pin,value); break;
            #endif                
            #ifdef SHIFTout
                case 'Y': 
                    if (value) {
                        xlOutputState |= (1UL<<pin);
                    } else {
                        xlOutputState &= ~(1UL<<pin);
                    }
                    write_XL_outputs(xlOutputState);                            
                    break;
            #endif
            default: strcpy_P(result,invalid_port); return;
        }
        strcpy_P(result,ok);
    }
}


//---------------------------------------------------------------------------------------------


// Execute an i command
//     i=read input

void command_i(char* result, const char* command) {
    size_t len=strlen(command);
    if (len==1) { // read all ports
        uint8_t pina=0;
        uint8_t pinb=0;
        uint8_t pinc=0;
        uint8_t pind=0;       
        #ifdef PORTA
            pina=in_port(PORTA,PINA);
        #endif
        #ifdef PORTB
            pinb=in_port(PORTB,PINB);
        #endif
        #ifdef PORTC
            pinc=in_port(PORTC,PINC);
         #endif
        #ifdef PORTD
            pind=in_port(PORTD,PIND);
        #endif
        // If the AVR has more than 4 ports, the read also ports E-H
        #ifdef PORTE
            uint8_t pine=0;
            uint8_t pinf=0;
            uint8_t ping=0;
            uint8_t pinh=0; 
            #ifdef PORTE
                pine=in_port(PORTE,PINE);
            #endif
            #ifdef PORTF
                pinf=in_port(PORTF,PINF);
            #endif
            #ifdef PORTG
                ping=in_port(PORTG,PING);
            #endif
            #ifdef PORTH
                pinh=in_port(PORTH,PINH);
            #endif
        #endif
        // If the controler has more than 4 ports, then output a 64bit number, else output a 32bit number
        #ifdef PORTE
            sprintf_P(result,PSTR("i=%02X%02X%02X%02X%02X%02X%02X%02X\n"),pinh,ping,pinf,pine,pind,pinc,pinb,pina);
        #else
           sprintf_P(result,PSTR("i=%02X%02X%02X%02X\n"),pind,pinc,pinb,pina);
        #endif
    }
    else if (len==3) {  // read a single port
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        uint8_t value;
        switch (port) {
            #ifdef PORTA
                case 'A': value=in_port(PORTA,PINA); break;
            #endif
            #ifdef PORTB
                case 'B': value=in_port(PORTB,PINB); break;
            #endif
            #ifdef PORTC
                case 'C': value=in_port(PORTC,PINC); break;
            #endif
            #ifdef PORTD
                case 'D': value=in_port(PORTD,PIND); break;
            #endif
            #ifdef PORTE
                case 'E': value=in_port(PORTE,PINE); break;
            #endif
            #ifdef PORTF
                case 'F': value=in_port(PORTF,PINF); break;
            #endif
            #ifdef PORTG
                case 'G': value=in_port(PORTG,PING); break;
            #endif
            #ifdef PORTH
                case 'H': value=in_port(PORTH,PINH); break;
            #endif
            #ifdef PORTI
                case 'I': value=in_port(PORTI,PINI); break;
            #endif
            #ifdef PORTJ
                case 'J': value=in_port(PORTJ,PINJ); break;
            #endif
            #ifdef PORTK
                case 'K': value=in_port(PORTK,PINK); break;
            #endif
            #ifdef PORTL
                case 'L': value=in_port(PORTL,PINL); break;
            #endif
            #ifdef SHIFTin
                case 'X': 
                  #if XL_INPUTS > 24                    
                    sprintf_P(result,PSTR("PX=%08lX\n"),read_XL_inputs()); return;
                  #elif XL_INPUTS > 16                    
                    sprintf_P(result,PSTR("PX=%06lX\n"),read_XL_inputs()); return;
                  #elif XL_INPUTS > 8           
                    sprintf_P(result,PSTR("PX=%04lX\n"),read_XL_inputs()); return;
                  #else                 
                    sprintf_P(result,PSTR("PX=%02lX\n"),read_XL_inputs()); return;
                  #endif
            #endif
            #ifdef SHIFTout
                case 'Y': 
                    #if XL_INPUTS > 24                    
                    sprintf_P(result,PSTR("PY=%08lX\n"),xlOutputState); return;
                  #elif XL_INPUTS > 16                    
                    sprintf_P(result,PSTR("PY=%06lX\n"),xlOutputState); return;
                  #elif XL_INPUTS > 8           
                    sprintf_P(result,PSTR("PY=%04lX\n"),xlOutputState); return;
                  #else                 
                    sprintf_P(result,PSTR("PY=%02lX\n"),xlOutputState); return;
                  #endif
            #endif
            default: strcpy_P(result,invalid_port); return;
        }
        sprintf_P(result,PSTR("P%c=%02X\n"),port,value);
    }
    else  {  // read a single port pin
        if (command[1]!='P') {
            strcpy_P(result,invalid_port);
            return;
        }
        char port=command[2];
        int pin=decodeDec(command+3);
        #ifdef SHIFTin
            if (port=='X') {
                if (pin<0 || pin>=XL_INPUTS) {
                    strcpy_P(result,invalid_pin);
                    return;
                }            
            }
            else 
        #endif
        #ifdef SHIFTout
            if (port=='Y') {
                if (pin<0 || pin>=XL_OUTPUTS) {
                    strcpy_P(result,invalid_pin);
                    return;
                }            
            }
            else 
        #endif
        if (pin<0 || pin>7) {
            strcpy_P(result,invalid_pin);
            return;
        }
        uint8_t value;
        switch (port) {
            #ifdef PORTA
                case 'A': value=in_pin(PORTA,PINA,pin); break;
            #endif
            #ifdef PORTB
                case 'B': value=in_pin(PORTB,PINB,pin); break;
            #endif
            #ifdef PORTC
                case 'C': value=in_pin(PORTC,PINC,pin); break;
            #endif
            #ifdef PORTD
                case 'D': value=in_pin(PORTD,PIND,pin); break;
            #endif
            #ifdef PORTE
                case 'E': value=in_pin(PORTE,PINE,pin); break;
            #endif
            #ifdef PORTF
                case 'F': value=in_pin(PORTF,PINF,pin); break;
            #endif
            #ifdef PORTG
                case 'G': value=in_pin(PORTG,PING,pin); break;
            #endif
            #ifdef PORTH
                case 'H': value=in_pin(PORTH,PINH,pin); break;
            #endif
            #ifdef PORTI
                case 'I': value=in_pin(PORTI,PINI,pin); break;
            #endif
            #ifdef PORTJ
                case 'J': value=in_pin(PORTJ,PINJ,pin); break;
            #endif
            #ifdef PORTK
                case 'K': value=in_pin(PORTK,PINK,pin); break;
            #endif
            #ifdef PORTL
                case 'L': value=in_pin(PORTL,PINL,pin); break;
            #endif
            #ifdef SHIFTin
                case 'X': value=readBit(read_XL_inputs(),pin); break;
            #endif
            #ifdef SHIFTout
                case 'Y': value=readBit(xlOutputState,pin); break;
            #endif                  
            default: strcpy_P(result,invalid_port); return;
        }
        sprintf_P(result,PSTR("P%c%i=%i\n"),port,pin,value);
    }
}

//--------------------------------------------------------------------------------------------

#if ADC_CHANNELS > 0

// Execute an a command
//     a=read analog input

void command_a(char* result, const char* command) {
      int channel=decodeDec(command+1);
      if (channel<0 || channel>=ADC_CHANNELS) {
          strcpy_P(result,invalid_channel);
          return;
      }
      sprintf_P(result,PSTR("ADC%i=%04X\n"),channel,getADC(channel));
}

// Execute an r command.
//     r=set ADC reference input
void command_r(char* result, const char* command) {
    if (setAREF(command+1)!=1) {
        strcpy_P(result,invalid_reference);
        return;
    }
    strcpy_P(result,ok);
}

#endif //ADC_CHANNELS


//-------------------------------------------------------------------------------------


// Execute an I/O command
// Return the result string in result buffer

void io_command(char* result, const char* command) {
    // call command function
    switch (command[0]) {
        // if the command is empty, return nothing
        case 0  : result[0]=0; break;
        // call the command function
        case 'd': command_d(result,command); break;
        case 'p': command_p(result,command); break;
        case 'o': command_o(result,command); break;
        case 'i': command_i(result,command); break;
        #if ADC_CHANNELS > 0
            case 'r': command_r(result,command); break;
            case 'a': command_a(result,command); break;
        #endif
        // in case of unknown command, return error message
        default: strcpy_P(result,bad_command);
    }
}
